@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 # copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 # contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
 #
@@ -32,44 +31,55 @@ __docformat__ = "restructuredtext en"
 
 import sys
 import os
-from os.path import (splitext, join, abspath, isdir, dirname, exists,
-                     basename, expanduser, normcase, realpath)
+from os.path import (
+    splitext,
+    join,
+    abspath,
+    isdir,
+    dirname,
+    exists,
+    expanduser,
+    normcase,
+    realpath,
+)
 from imp import find_module, load_module, C_BUILTIN, PY_COMPILED, PKG_DIRECTORY
-from distutils.sysconfig import get_config_var, get_python_lib, get_python_version
+from distutils.sysconfig import get_python_lib
 from distutils.errors import DistutilsPlatformError
+from typing import Dict, List, Optional, Any, Tuple, Union, Sequence
+from types import ModuleType
+from _frozen_importlib_external import FileFinder
 
-from six import PY3
-from six.moves import map, range
+from logilab.common import STD_BLACKLIST, _handle_blacklist
+from logilab.common.deprecation import callable_deprecated
 
 try:
     import zipimport
 except ImportError:
-    zipimport = None
+    # mypy: Incompatible types in assignment (expression has type "None", variable has type Module)
+    # conditional import
+    zipimport = None  # type: ignore
 
 ZIPFILE = object()
 
-from logilab.common import STD_BLACKLIST, _handle_blacklist
-from logilab.common.deprecation import deprecated
-
 # Notes about STD_LIB_DIR
 # Consider arch-specific installation for STD_LIB_DIR definition
 # :mod:`distutils.sysconfig` contains to much hardcoded values to rely on
 #
 # :see: `Problems with /usr/lib64 builds <http://bugs.python.org/issue1294959>`_
 # :see: `FHS <http://www.pathname.com/fhs/pub/fhs-2.3.html#LIBLTQUALGTALTERNATEFORMATESSENTIAL>`_
-if sys.platform.startswith('win'):
-    PY_SOURCE_EXTS = ('py', 'pyw')
-    PY_COMPILED_EXTS = ('dll', 'pyd')
+if sys.platform.startswith("win"):
+    PY_SOURCE_EXTS = ("py", "pyw")
+    PY_COMPILED_EXTS = ("dll", "pyd")
 else:
-    PY_SOURCE_EXTS = ('py',)
-    PY_COMPILED_EXTS = ('so',)
+    PY_SOURCE_EXTS = ("py",)
+    PY_COMPILED_EXTS = ("so",)
 
 try:
     STD_LIB_DIR = get_python_lib(standard_lib=True)
 # get_python_lib(standard_lib=1) is not available on pypy, set STD_LIB_DIR to
 # non-valid path, see https://bugs.pypy.org/issue1164
 except DistutilsPlatformError:
-    STD_LIB_DIR = '//'
+    STD_LIB_DIR = "//"
 
 EXT_LIB_DIR = get_python_lib()
 
@@ -81,7 +91,19 @@ class NoSourceFile(Exception):
     source file for a precompiled file
     """
 
+
 class LazyObject(object):
+    """
+    This class allows to lazyly declare a object (most likely only a callable
+    according to the code) from a module without importing it.
+
+    The import will be triggered when the user tries to access attributes of
+    the object/callable or call it.
+
+    Trying to set or delete attributes of the wrapped object/callable will not
+    works as expected.
+    """
+
     def __init__(self, module, obj):
         self.module = module
         self.obj = obj
@@ -89,21 +111,22 @@ class LazyObject(object):
 
     def _getobj(self):
         if self._imported is None:
-           self._imported = getattr(load_module_from_name(self.module),
-                                    self.obj)
+            self._imported = getattr(load_module_from_name(self.module), self.obj)
         return self._imported
 
     def __getattribute__(self, attr):
         try:
             return super(LazyObject, self).__getattribute__(attr)
-        except AttributeError as ex:
+        except AttributeError:
             return getattr(self._getobj(), attr)
 
     def __call__(self, *args, **kwargs):
         return self._getobj()(*args, **kwargs)
 
 
-def load_module_from_name(dotted_name, path=None, use_sys=True):
+def load_module_from_name(
+    dotted_name: str, path: Optional[Any] = None, use_sys: int = True
+) -> ModuleType:
     """Load a Python module from its name.
 
     :type dotted_name: str
@@ -125,10 +148,15 @@ def load_module_from_name(dotted_name, path=None, use_sys=True):
     :rtype: module
     :return: the loaded module
     """
-    return load_module_from_modpath(dotted_name.split('.'), path, use_sys)
+    module = load_module_from_modpath(dotted_name.split("."), path, use_sys)
+    if module is None:
+        raise ImportError("module %s doesn't exist" % dotted_name)
+    return module
 
 
-def load_module_from_modpath(parts, path=None, use_sys=True):
+def load_module_from_modpath(
+    parts: List[str], path: Optional[Any] = None, use_sys: int = True
+) -> Optional[ModuleType]:
     """Load a python module from its splitted name.
 
     :type parts: list(str) or tuple(str)
@@ -151,14 +179,14 @@ def load_module_from_modpath(parts, path=None, use_sys=True):
     """
     if use_sys:
         try:
-            return sys.modules['.'.join(parts)]
+            return sys.modules[".".join(parts)]
         except KeyError:
             pass
     modpath = []
     prevmodule = None
     for part in parts:
         modpath.append(part)
-        curname = '.'.join(modpath)
+        curname = ".".join(modpath)
         module = None
         if len(modpath) != len(parts):
             # even with use_sys=False, should try to get outer packages from sys.modules
@@ -169,19 +197,22 @@ def load_module_from_modpath(parts, path=None, use_sys=True):
         if module is None:
             mp_file, mp_filename, mp_desc = find_module(part, path)
             try:
-                module = load_module(curname, mp_file, mp_filename, mp_desc)
+                # mypy: Argument 2 to "load_module" has incompatible type "IO[Any]";
+                # mypy: expected "Optional[_FileLike]"
+                # this is handled by the try/finally somehow?
+                module = load_module(curname, mp_file, mp_filename, mp_desc)  # type: ignore
             finally:
                 if mp_file is not None:
                     mp_file.close()
         if prevmodule:
             setattr(prevmodule, part, module)
-        _file = getattr(module, '__file__', '')
+        _file = getattr(module, "__file__", "")
         prevmodule = module
         if not _file and _is_namespace(curname):
             continue
         if not _file and len(modpath) != len(parts):
-            raise ImportError('no module in %s' % '.'.join(parts[len(modpath):]) )
-        path = [dirname( _file )]
+            raise ImportError("no module in %s" % ".".join(parts[len(modpath) :]))
+        path = [dirname(_file)]
     return module
 
 
@@ -211,32 +242,23 @@ def load_module_from_file(filepath, path=None, use_sys=True, extrapath=None):
     return load_module_from_modpath(modpath, path, use_sys)
 
 
-def _check_init(path, mod_path):
+def _check_init(path: str, mod_path: List[str]) -> bool:
     """check there are some __init__.py all along the way"""
     modpath = []
     for part in mod_path:
         modpath.append(part)
         path = join(path, part)
-        if not _is_namespace('.'.join(modpath)) and not _has_init(path):
+        if not _is_namespace(".".join(modpath)) and not _has_init(path):
             return False
     return True
 
 
-def _canonicalize_path(path):
+def _canonicalize_path(path: str) -> str:
     return realpath(expanduser(path))
 
 
-def _path_from_filename(filename):
-    if PY3:
-        return filename
-    else:
-        if filename.endswith(".pyc"):
-            return filename[:-1]
-        return filename
-
-
-@deprecated('you should avoid using modpath_from_file()')
-def modpath_from_file(filename, extrapath=None):
+@callable_deprecated("you should avoid using modpath_from_file()")
+def modpath_from_file(filename: str, extrapath: Optional[Dict[str, str]] = None) -> List[str]:
     """DEPRECATED: doens't play well with symlinks and sys.meta_path
 
     Given a file path return the corresponding splitted module's name
@@ -258,30 +280,29 @@ def modpath_from_file(filename, extrapath=None):
     :rtype: list(str)
     :return: the corresponding splitted module's name
     """
-    filename = _path_from_filename(filename)
     filename = _canonicalize_path(filename)
     base = os.path.splitext(filename)[0]
 
     if extrapath is not None:
         for path_ in map(_canonicalize_path, extrapath):
             path = abspath(path_)
-            if path and normcase(base[:len(path)]) == normcase(path):
-                submodpath = [pkg for pkg in base[len(path):].split(os.sep)
-                              if pkg]
+            if path and normcase(base[: len(path)]) == normcase(path):
+                submodpath = [pkg for pkg in base[len(path) :].split(os.sep) if pkg]
                 if _check_init(path, submodpath[:-1]):
-                    return extrapath[path_].split('.') + submodpath
+                    return extrapath[path_].split(".") + submodpath
 
     for path in map(_canonicalize_path, sys.path):
         if path and normcase(base).startswith(path):
-            modpath = [pkg for pkg in base[len(path):].split(os.sep) if pkg]
+            modpath = [pkg for pkg in base[len(path) :].split(os.sep) if pkg]
             if _check_init(path, modpath[:-1]):
                 return modpath
 
-    raise ImportError('Unable to find module for %s in %s' % (
-        filename, ', \n'.join(sys.path)))
+    raise ImportError("Unable to find module for %s in %s" % (filename, ", \n".join(sys.path)))
 
 
-def file_from_modpath(modpath, path=None, context_file=None):
+def file_from_modpath(
+    modpath: List[str], path: Optional[Any] = None, context_file: Optional[str] = None
+) -> Optional[str]:
     """given a mod path (i.e. splitted module / package name), return the
     corresponding file, giving priority to source file over precompiled
     file if it exists
@@ -311,24 +332,24 @@ def file_from_modpath(modpath, path=None, context_file=None):
       the path to the module's file or None if it's an integrated
       builtin module such as 'sys'
     """
+    context: Optional[str]
     if context_file is not None:
         context = dirname(context_file)
     else:
         context = context_file
-    if modpath[0] == 'xml':
+    if modpath[0] == "xml":
         # handle _xmlplus
         try:
-            return _file_from_modpath(['_xmlplus'] + modpath[1:], path, context)
+            return _file_from_modpath(["_xmlplus"] + modpath[1:], path, context)
         except ImportError:
             return _file_from_modpath(modpath, path, context)
-    elif modpath == ['os', 'path']:
+    elif modpath == ["os", "path"]:
         # FIXME: currently ignoring search_path...
         return os.path.__file__
     return _file_from_modpath(modpath, path, context)
 
 
-
-def get_module_part(dotted_name, context_file=None):
+def get_module_part(dotted_name: str, context_file: Optional[str] = None) -> str:
     """given a dotted name return the module part of the name :
 
     >>> get_module_part('logilab.common.modutils.get_module_part')
@@ -355,9 +376,9 @@ def get_module_part(dotted_name, context_file=None):
     (see #10066)
     """
     # os.path trick
-    if dotted_name.startswith('os.path'):
-        return 'os.path'
-    parts = dotted_name.split('.')
+    if dotted_name.startswith("os.path"):
+        return "os.path"
+    parts = dotted_name.split(".")
     if context_file is not None:
         # first check for builtin module which won't be considered latter
         # in that case (path != None)
@@ -366,28 +387,29 @@ def get_module_part(dotted_name, context_file=None):
                 raise ImportError(dotted_name)
             return parts[0]
         # don't use += or insert, we want a new list to be created !
-    path = None
+    path: Optional[List] = None
     starti = 0
-    if parts[0] == '':
-        assert context_file is not None, \
-                'explicit relative import, but no context_file?'
-        path = [] # prevent resolving the import non-relatively
+    if parts[0] == "":
+        assert context_file is not None, "explicit relative import, but no context_file?"
+        path = []  # prevent resolving the import non-relatively
         starti = 1
-    while parts[starti] == '': # for all further dots: change context
+    while parts[starti] == "":  # for all further dots: change context
         starti += 1
+        assert context_file is not None
         context_file = dirname(context_file)
     for i in range(starti, len(parts)):
         try:
-            file_from_modpath(parts[starti:i+1],
-                    path=path, context_file=context_file)
+            file_from_modpath(parts[starti : i + 1], path=path, context_file=context_file)
         except ImportError:
             if not i >= max(1, len(parts) - 2):
                 raise
-            return '.'.join(parts[:i])
+            return ".".join(parts[:i])
     return dotted_name
 
 
-def get_modules(package, src_directory, blacklist=STD_BLACKLIST):
+def get_modules(
+    package: str, src_directory: str, blacklist: Sequence[str] = STD_BLACKLIST
+) -> List[str]:
     """given a package directory return a list of all available python
     modules in the package and its subpackages
 
@@ -412,22 +434,21 @@ def get_modules(package, src_directory, blacklist=STD_BLACKLIST):
     for directory, dirnames, filenames in os.walk(src_directory):
         _handle_blacklist(blacklist, dirnames, filenames)
         # check for __init__.py
-        if not '__init__.py' in filenames:
+        if "__init__.py" not in filenames:
             dirnames[:] = ()
             continue
         if directory != src_directory:
-            dir_package = directory[len(src_directory):].replace(os.sep, '.')
+            dir_package = directory[len(src_directory) :].replace(os.sep, ".")
             modules.append(package + dir_package)
         for filename in filenames:
-            if _is_python_file(filename) and filename != '__init__.py':
+            if _is_python_file(filename) and filename != "__init__.py":
                 src = join(directory, filename)
-                module = package + src[len(src_directory):-3]
-                modules.append(module.replace(os.sep, '.'))
+                module = package + src[len(src_directory) : -3]
+                modules.append(module.replace(os.sep, "."))
     return modules
 
 
-
-def get_module_files(src_directory, blacklist=STD_BLACKLIST):
+def get_module_files(src_directory: str, blacklist: Sequence[str] = STD_BLACKLIST) -> List[str]:
     """given a package directory return a list of all available python
     module's files in the package and its subpackages
 
@@ -449,7 +470,7 @@ def get_module_files(src_directory, blacklist=STD_BLACKLIST):
     for directory, dirnames, filenames in os.walk(src_directory):
         _handle_blacklist(blacklist, dirnames, filenames)
         # check for __init__.py
-        if not '__init__.py' in filenames:
+        if "__init__.py" not in filenames:
             dirnames[:] = ()
             continue
         for filename in filenames:
@@ -459,7 +480,7 @@ def get_module_files(src_directory, blacklist=STD_BLACKLIST):
     return files
 
 
-def get_source_file(filename, include_no_ext=False):
+def get_source_file(filename: str, include_no_ext: bool = False) -> str:
     """given a python module's file name return the matching source file
     name (the filename will be returned identically if it's a already an
     absolute path to a python source file...)
@@ -475,7 +496,7 @@ def get_source_file(filename, include_no_ext=False):
     """
     base, orig_ext = splitext(abspath(filename))
     for ext in PY_SOURCE_EXTS:
-        source_path = '%s.%s' % (base, ext)
+        source_path = "%s.%s" % (base, ext)
         if exists(source_path):
             return source_path
     if include_no_ext and not orig_ext and exists(base):
@@ -487,7 +508,7 @@ def cleanup_sys_modules(directories):
     """remove submodules of `directories` from `sys.modules`"""
     cleaned = []
     for modname, module in list(sys.modules.items()):
-        modfile = getattr(module, '__file__', None)
+        modfile = getattr(module, "__file__", None)
         if modfile:
             for directory in directories:
                 if modfile.startswith(directory):
@@ -517,7 +538,9 @@ def is_python_source(filename):
     return splitext(filename)[1][1:] in PY_SOURCE_EXTS
 
 
-def is_standard_module(modname, std_path=(STD_LIB_DIR,)):
+def is_standard_module(
+    modname: str, std_path: Union[List[str], Tuple[str]] = (STD_LIB_DIR,)
+) -> bool:
     """try to guess if a module is a standard python module (by default,
     see `std_path` parameter's description)
 
@@ -537,10 +560,10 @@ def is_standard_module(modname, std_path=(STD_LIB_DIR,)):
     Note: this function is known to return wrong values when inside virtualenv.
     See https://www.logilab.org/ticket/294756.
     """
-    modname = modname.split('.')[0]
+    modname = modname.split(".")[0]
     try:
         filename = file_from_modpath([modname])
-    except ImportError as ex:
+    except ImportError:
         # import failed, i'm probably not so wrong by supposing it's
         # not standard...
         return False
@@ -558,8 +581,7 @@ def is_standard_module(modname, std_path=(STD_LIB_DIR,)):
     return False
 
 
-
-def is_relative(modname, from_file):
+def is_relative(modname: str, from_file: str) -> bool:
     """return true if the given module name is relative to the given
     file name
 
@@ -579,7 +601,7 @@ def is_relative(modname, from_file):
     if from_file in sys.path:
         return False
     try:
-        find_module(modname.split('.')[0], [from_file])
+        find_module(modname.split(".")[0], [from_file])
         return True
     except ImportError:
         return False
@@ -587,7 +609,10 @@ def is_relative(modname, from_file):
 
 # internal only functions #####################################################
 
-def _file_from_modpath(modpath, path=None, context=None):
+
+def _file_from_modpath(
+    modpath: List[str], path: Optional[Any] = None, context: Optional[str] = None
+) -> Optional[str]:
     """given a mod path (i.e. splitted module / package name), return the
     corresponding file
 
@@ -604,6 +629,7 @@ def _file_from_modpath(modpath, path=None, context=None):
         mtype, mp_filename = _module_file(modpath, path)
     if mtype == PY_COMPILED:
         try:
+            assert mp_filename is not None
             return get_source_file(mp_filename)
         except NoSourceFile:
             return mp_filename
@@ -611,31 +637,44 @@ def _file_from_modpath(modpath, path=None, context=None):
         # integrated builtin module
         return None
     elif mtype == PKG_DIRECTORY:
+        assert mp_filename is not None
         mp_filename = _has_init(mp_filename)
     return mp_filename
 
-def _search_zip(modpath, pic):
+
+def _search_zip(
+    modpath: List[str], pic: Dict[str, Optional[FileFinder]]
+) -> Tuple[object, str, str]:
     for filepath, importer in pic.items():
         if importer is not None:
             if importer.find_module(modpath[0]):
-                if not importer.find_module('/'.join(modpath)):
-                    raise ImportError('No module named %s in %s/%s' % (
-                        '.'.join(modpath[1:]), filepath, modpath))
-                return ZIPFILE, abspath(filepath) + '/' + '/'.join(modpath), filepath
-    raise ImportError('No module named %s' % '.'.join(modpath))
+                if not importer.find_module("/".join(modpath)):
+                    raise ImportError(
+                        "No module named %s in %s/%s" % (".".join(modpath[1:]), filepath, modpath)
+                    )
+                return ZIPFILE, abspath(filepath) + "/" + "/".join(modpath), filepath
+    raise ImportError("No module named %s" % ".".join(modpath))
+
 
 try:
     import pkg_resources
 except ImportError:
-    pkg_resources = None
+    # mypy: Incompatible types in assignment (expression has type "None", variable has type Module)
+    # conditional import
+    pkg_resources = None  # type: ignore
 
 
-def _is_namespace(modname):
-    return (pkg_resources is not None
-            and modname in pkg_resources._namespace_packages)
+def _is_namespace(modname: str) -> bool:
+    # mypy: Module has no attribute "_namespace_packages"; maybe "fixup_namespace_packages"?"
+    # but is still has? or is it a failure from python3 port?
+    return (
+        pkg_resources is not None and modname in pkg_resources._namespace_packages  # type: ignore
+    )
 
 
-def _module_file(modpath, path=None):
+def _module_file(
+    modpath: List[str], path: Optional[List[str]] = None
+) -> Tuple[Union[int, object], Optional[str]]:
     """get a module type / file path
 
     :type modpath: list or tuple
@@ -655,9 +694,9 @@ def _module_file(modpath, path=None):
     # egg support compat
     try:
         pic = sys.path_importer_cache
-        _path = (path is None and sys.path or path)
+        _path = path if path is not None else sys.path
         for __path in _path:
-            if not __path in pic:
+            if __path not in pic:
                 try:
                     pic[__path] = zipimport.zipimporter(__path)
                 except zipimport.ZipImportError:
@@ -666,16 +705,20 @@ def _module_file(modpath, path=None):
     except AttributeError:
         checkeggs = False
     # pkg_resources support (aka setuptools namespace packages)
-    if (_is_namespace(modpath[0]) and modpath[0] in sys.modules):
+    if _is_namespace(modpath[0]) and modpath[0] in sys.modules:
         # setuptools has added into sys.modules a module object with proper
         # __path__, get back information from there
         module = sys.modules[modpath.pop(0)]
         # use list() to protect against _NamespacePath instance we get with python 3, which
         # find_module later doesn't like
-        path = list(module.__path__)
+        # mypy: Module has no attribute "__path__"
+        # I guess it does thanks to logilab's magic?
+        path = list(module.__path__)  # type: ignore
         if not modpath:
             return C_BUILTIN, None
+
     imported = []
+
     while modpath:
         modname = modpath[0]
         # take care to changes in find_module implementation wrt builtin modules
@@ -712,42 +755,43 @@ def _module_file(modpath, path=None):
         mtype = mp_desc[2]
         if modpath:
             if mtype != PKG_DIRECTORY:
-                raise ImportError('No module %s in %s' % ('.'.join(modpath),
-                                                          '.'.join(imported)))
+                raise ImportError("No module %s in %s" % (".".join(modpath), ".".join(imported)))
             # XXX guess if package is using pkgutil.extend_path by looking for
             # those keywords in the first four Kbytes
             try:
-                with open(join(mp_filename, '__init__.py')) as stream:
+                with open(join(mp_filename, "__init__.py")) as stream:
                     data = stream.read(4096)
             except IOError:
                 path = [mp_filename]
             else:
-                if 'pkgutil' in data and 'extend_path' in data:
+                if "pkgutil" in data and "extend_path" in data:
                     # extend_path is called, search sys.path for module/packages
                     # of this name see pkgutil.extend_path documentation
-                    path = [join(p, *imported) for p in sys.path
-                            if isdir(join(p, *imported))]
+                    path = [join(p, *imported) for p in sys.path if isdir(join(p, *imported))]
                 else:
                     path = [mp_filename]
     return mtype, mp_filename
 
-def _is_python_file(filename):
+
+def _is_python_file(filename: str) -> bool:
     """return true if the given filename should be considered as a python file
 
     .pyc and .pyo are ignored
     """
-    for ext in ('.py', '.so', '.pyd', '.pyw'):
+    for ext in (".py", ".so", ".pyd", ".pyw"):
         if filename.endswith(ext):
             return True
     return False
 
 
-def _has_init(directory):
+def _has_init(directory: str) -> Optional[str]:
     """if the given directory has a valid __init__ file, return its path,
     else return None
     """
-    mod_or_pack = join(directory, '__init__')
-    for ext in PY_SOURCE_EXTS + ('pyc', 'pyo'):
-        if exists(mod_or_pack + '.' + ext):
-            return mod_or_pack + '.' + ext
+    mod_or_pack = join(directory, "__init__")
+
+    for ext in PY_SOURCE_EXTS + ("pyc", "pyo"):
+        if exists(mod_or_pack + "." + ext):
+            return mod_or_pack + "." + ext
+
     return None
